import logging
import re
import os
from . import download_files_utilities as utils
from .file_regex import maven_kp_l2_regex
import numpy as np
import collections


# The code in this file was originally part of read_iuvs_file.py. According to
# justin.deighan@lasp.colorado.edu, these modes will not be supported with KP data.
# We'll keep the code alive here just in case they change their minds in the future.

def read_iuvs_file_unused_modes(file):
    """
    Read an IUVS file and return a dictionary containing the data.

    Parameters:
        file (str): The path to the IUVS file to be read.

    Returns:
        dict: A dictionary containing the data read from the IUVS file.
    """
    iuvs_dict = {}
    periapse_num = 0
    occ_num = 0
    with open(file) as f:
        line = f.readline()
        while line != "":
            if line.startswith("*"):
                # Read the header
                line = f.readline()
                obs_mode = line[19 : len(line) - 1].strip()

                header = {}
                f.readline()
                line = f.readline()
                header["time_start"] = line[19 : len(line) - 1].strip()
                line = f.readline()
                header["time_stop"] = line[19 : len(line) - 1].strip()
                line = f.readline()
                if obs_mode == "OCCULTATION":
                    header["target_name"] = line[19 : len(line) - 1].strip()
                    line = f.readline()
                header["sza"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["local_time"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lat_mso"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lon_mso"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["orbit_number"] = int(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["mars_season"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_mso_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_mso_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["subsol_geo_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["subsol_geo_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_sza"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_local_time"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_alt"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["mars_sun_dist"] = float(line[19 : len(line) - 1].strip())

                if obs_mode == "CORONA_LORES_HIGH":
                    line = f.readline()
                    n_alt_bins = int(line[19 : len(line) - 1].strip())
                    header["n_alt_bins"] = float(n_alt_bins)

                    iuvs_dict["corona_lores_high"] = {}
                    iuvs_dict["corona_lores_high"].update(header)

                    f.readline()

                    # Read the Half int
                    line = f.readline()
                    half_int_dist_labels = line[19 : len(line) - 1].strip().split()
                    half_int_dist = collections.OrderedDict(
                        (x, []) for x in half_int_dist_labels
                    )
                    half_int_dist_unc = collections.OrderedDict(
                        (x, []) for x in half_int_dist_labels
                    )
                    line = f.readline()
                    vals = line[26 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        half_int_dist[list(half_int_dist.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[26 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        half_int_dist_unc[list(half_int_dist_unc.keys())[index]].append(
                            val
                        )
                        index += 1

                    iuvs_dict["corona_lores_high"]["half_int_dist"] = half_int_dist
                    iuvs_dict["corona_lores_high"][
                        "half_int_dist_unc"
                    ] = half_int_dist_unc

                    # Blank space
                    f.readline()
                    f.readline()

                    # Read in the density
                    line = f.readline()
                    density_labels = line.strip().split()
                    density = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            density[list(density.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["corona_lores_high"]["density"] = density

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density systematic uncertainty
                    density_sys_unc = collections.OrderedDict(
                        (x, []) for x in density_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        density_sys_unc[list(density.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["corona_lores_high"]["density_sys_unc"] = density_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density uncertainty
                    density_unc = collections.OrderedDict(
                        (x, []) for x in density_labels
                    )
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            density_unc[list(density.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["corona_lores_high"]["density_unc"] = density_unc

                    f.readline()
                    f.readline()

                    line = f.readline()
                    radiance_labels = line.strip().split()
                    if "Cameron" in radiance_labels:
                        radiance_labels.remove("Cameron")
                    radiance = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            radiance[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["corona_lores_high"]["radiance"] = radiance

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    radiance_sys_unc = collections.OrderedDict(
                        (x, []) for x in radiance_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["corona_lores_high"][
                        "radiance_sys_unc"
                    ] = radiance_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance uncertainty
                    radiance_unc = collections.OrderedDict(
                        (x, []) for x in radiance_labels
                    )
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            radiance_unc[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["corona_lores_high"]["radiance_unc"] = radiance_unc

                elif obs_mode == "APOAPSE":

                    f.readline()
                    maps = {}
                    for j in range(0, 17):
                        var = f.readline().strip()
                        line = f.readline()
                        lons = line.strip().split()
                        lons = [float(x) for x in lons]
                        lats = []
                        data = []
                        for k in range(0, 45):
                            line = f.readline().strip().split()
                            lats.append(float(line[0]))
                            line_data = line[1:]
                            line_data = [
                                float(x) if x != "-9.9999990E+09" else float("nan")
                                for x in line_data
                            ]
                            data.append(line_data)

                        maps[var] = data
                        f.readline()

                    maps["latitude"] = lats
                    maps["longitude"] = lons

                    iuvs_dict["apoapse"] = {}
                    iuvs_dict["apoapse"].update(header)
                    iuvs_dict["apoapse"].update(maps)

                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    line = f.readline()
                    radiance_labels = line.strip().split()
                    radiance_sys_unc = collections.OrderedDict(
                        (x, []) for x in radiance_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["apoapse"]["radiance_sys_unc"] = radiance_sys_unc

            line = f.readline()

    return iuvs_dict





from .load import load
from pyspedas.utilities.datasets import find_datasets

from .threedp import threedp
from .orbit import orbit
from .sms import sms
from .waves import waves
from .mfi import mfi
from .swe import swe
from .datasets import datasets



from .load import load
from .het import het
from .let import let
from .sit import sit
from .sept import sept
from .ste import ste
from .swea import swea
from .mag import mag
from .plastic import plastic
from .waves import waves
from .beacon import beacon
from .datasets import datasets



import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(trange=['2006-06-01', '2006-06-02'], 
         instrument='celias',
         datatype='pm_5min',
         suffix='', 
         prefix='',
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the SOHO mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.projects.soho.celias
        pyspedas.projects.soho.cosp
        pyspedas.projects.soho.erne
        pyspedas.projects.soho.orbit

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2006-06-01', '2006-06-02']

        instrument: str
            Spacecraft identifier ('celias', 'cosp', 'erne', 'orbit')
            Default: 'celias'

        datatype: str
            Valid options: 'pm_5min'
            Default: 'pm_5min'

        suffix: str
            The tplot variable names will be given this suffix.
            no prefix is added.
            Default: ''
        
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: 'False', only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            Default: [], all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False
        
        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

    Returns
    ----------
        List of tplot variables created.

    Example
    ----------
        import pyspedas
        from pyspedas.tplot_tools import tplot
        celias_soho_vars = pyspedas.projects.soho.celias(trange=['2006-06-01', '2006-06-02'])

        cosp_soho_vars = pyspedas.projects.soho.cosp(trange=['2006-06-01', '2006-06-02'])

        erne_soho_vars = pyspedas.projects.soho.erne(trange=['2006-06-01', '2006-06-02'])

        orbit_soho_vars = pyspedas.projects.soho.orbit(trange=['2006-06-01', '2006-06-02'])

    """

    if prefix is None:
        prefix = ''
    
    if suffix is None:
        suffix = ''

    res = 24 * 3600.

    if instrument == 'celias':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'costep':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y0101_v??.??.cdf'
        res = 24 * 3600. * 366
    elif instrument == 'erne':
        pathformat = instrument+'/'+datatype+'/%Y/soho_'+instrument+'-'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'orbit':
        if datatype not in ['pre_or', 'def_or', 'def_at']:
            logging.error('Invalod datatype: ' + datatype)
            return
        datatype_fn = datatype.split('_')[1] + '_' +datatype.split('_')[0]
        pathformat = instrument+'/'+datatype+'/cdf/%Y/so_'+datatype_fn+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange, res=res)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


from .load import load
from pyspedas.utilities.datasets import find_datasets
from .load_csa import load_csa
from typing import List, Union, Optional

from .fgm import fgm
from .aspoc import aspoc
from .cis import cis
from .dwp import dwp
from .edi import edi
from .efw import efw
from .peace import peace
from .rapid import rapid
from .staff import staff
from .wbd import wbd
from .whi import whi
from .datasets import datasets


import logging
import numpy as np
from pyspedas.tplot_tools import get_data, store_data, options
from pyspedas.tplot_tools import tnames

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean


def rbsp_rbspice_omni(probe='a', datatype='TOFxEH', level='l3'):
    """
    Calculates the omni-directional flux for all 6 telescopes
    
    Parameters
    ----------
    probe : str or list of str, default='a'
        Spacecraft probe name: 'a' or 'b'

    datatype: str, default='TOFxEH'
        RBSPICE data type::

         'EBR', 'ESRHELT', 'ESRLEHT', 'IBR', 'ISBR', 'ISRHELT',
         'TOFxEH', 'TOFxEIon', 'TOFxEnonH', 'TOFxPHHHELT', 'TOFxPHHLEHT'

        Values depends on different data levels.

    level : str, default='l3'
        data level: 'l1','l2','l3'

    Returns
    -------
    out : list
        Tplot variables created

    Examples
    --------
    This function is called within pyspedas.projects.rbsp.rbspice
    """
    if probe is None:
        probe = 'a'
    if datatype is None:
        datatype = 'TOFxEH'
    if level is None:
        level = 'l3'
    if level != 'l1':
        units_label = '1/(cm^2-sr-s-keV)'
    else:
        units_label = 'counts/s'

    prefix = 'rbsp'+probe+'_rbspice_'+level+'_'+datatype+'_'
    
    # find the flux/cps data name(s)
    data_var = tnames(prefix + 'F*DU')
    
    if not data_var:
        logging.error('Error, problem finding the RBSPICE data to calculate omni-directional spectrograms')
        return

    logging.info('Calculating omni directional energy spectra; this might take a few minutes...')
    out = []

    for i in range(len(data_var)):
        species_str = data_var[i][-4:-2]
        if species_str == 'FP':
            species='proton'
            if datatype != 'TOFxPHHHELT':
                zrange = [5., 1.e5]
            else:
                zrange = [2.e2, 1.e6]
        elif species_str == 'He':
            species = 'helium'
            zrange = [1., 5.e2]
        elif species_str == 'FO':
            species = 'oxygen'
            if datatype != 'TOFxPHHHELT':
                zrange = [1., 1.e2]
            else:
                zrange = [1e1, 1.e4]
        
        # load the flux/cps data
        d = get_data(prefix+species)
        d_for_en_table = get_data(prefix+species+'_T0')
        
        if d is not None:
            flux_omni = np.zeros((len(d.times),len(d.y[0, :, 0])))
            for k in range(len(d.times)):
                for l in range(len(d.y[0, :, 0])):
                    flux_omni[k, l] = nanmean(d.y[k, l, :])
            newname = prefix+species+'_omni'
            store_data(newname, data={'x': d.times, 'y': flux_omni, 'v': d_for_en_table.v})
            options(newname, 'ylog', True)
            options(newname, 'zlog', True)
            options(newname, 'spec', True)
            options(newname, 'zrange', zrange)
            options(newname, 'ytitle', 'rbsp-'+probe+'\nrbspice\n'+species+'\nomni')
            options(newname, 'ysubtitle', '[keV]')
            options(newname, 'ztitle', units_label)
            out.append(newname)
    return out



from .load import load

# This routine was moved out of __init__.py.  Please see that file for previous revision history.


def lad(trange=['2018-11-5', '2018-11-6'], 
        probe='1',
        datatype='',
        prefix='',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        force_download=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the LAD instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        datatype: str
            Data type; Valid options: ''
            Default: ''

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: '' (all variables loaded)

        varnames: list of str
            List of variable names to load
            Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables
            Default: False

        force_download: bool
            Set this flag to download the CDF files, even if the local copy is newer
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    --------
        list of str
            List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> lad_vars = pyspedas.projects.twins.lad(trange=['2018-11-5/6:00', '2018-11-5/6:20'], time_clip=True)
    >>> tplot(['lad1_data', 'lad2_data'])

    """
    return load(instrument='lad', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, force_download=force_download, notplot=notplot, time_clip=time_clip, no_update=no_update)




import unittest
from pyspedas import CDAWeb, data_exists, del_data, get_data, time_double
from pyspedas.cdagui_tools.config import CONFIG

output_dir = CONFIG["local_data_dir"]


class CDAWebTests(unittest.TestCase):
    def test_get_observatories(self):
        cdaweb_obj = CDAWeb()
        obslist = cdaweb_obj.get_observatories()
        self.assertTrue("ARTEMIS" in obslist)

    def test_get_instruments(self):
        cdaweb_obj = CDAWeb()
        inst_list = cdaweb_obj.get_instruments()
        self.assertTrue("Magnetic Fields (space)" in inst_list)

    def test_get_datasets(self):
        cdaweb_obj = CDAWeb()
        ds_list = cdaweb_obj.get_datasets(["Voyager"], ["Magnetic Fields (space)"])
        self.assertTrue("VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)" in ds_list)

    def test_get_filenames(self):
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(
            ["VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)"], "1979-01-11", "1979-01-12"
        )
        self.assertTrue(
            any(["voyager2_2s_mag_19790111_v01.cdf" in item for item in urllist])
        )

    def test_load_merge(self):
        del_data("*")
        # Create the CDAWeb interface object
        cdaweb_obj = CDAWeb()

        # We'll pick one of available data sets and load it into tplot variables
        dataset = "VOYAGER2_COHO1HR_MERGED_MAG_PLASMA"
        start_time = "2009-01-01 00:00:00"
        end_time = "2010-01-01 00:00:00"

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)

        # Download the data and load as tplot variables.  Setting a prefix
        # is useful if you want to work with both Voyager 1 and Voyager 2
        # data; the variable names in the archived data are the same for both
        # spacecraft.

        cdaweb_obj.cda_download(urllist, output_dir, prefix="v2_")
        dat = get_data("v2_protonDensity")
        # Files are monthly; we should have over a year of data.
        self.assertTrue((dat.times[-1] - dat.times[0]) / 86400.0 > 395.0)

    def test_load_icon_netcdf(self):
        del_data("*")
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = "2021-01-15T14:05:52"
        time1 = "2021-01-15T15:08:57"

        urllist = cdaweb_obj.get_filenames(
            ["ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)"],
            time0,
            time1,
        )
        result = cdaweb_obj.cda_download(urllist, output_dir)
        self.assertEqual(len(result), 3)
        self.assertTrue(data_exists("ICON_L22_Fringe_Amplitude"))

    def test_load_icon_netcdf_default_dir(self):
        del_data("*")
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = "2021-01-15T14:05:52"
        time1 = "2021-01-15T15:08:57"

        urllist = cdaweb_obj.get_filenames(
            ["ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)"],
            time0,
            time1,
        )
        result = cdaweb_obj.cda_download(urllist, output_dir)
        self.assertEqual(len(result) > 0, True)
        self.assertTrue(data_exists("ICON_L22_Fringe_Amplitude"))

    def test_load_goes_netcdf(self):
        del_data("*")
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = "2021-01-15T14:05:52"
        time1 = "2021-01-15T15:08:57"

        urllist = cdaweb_obj.get_filenames(
            ["DN_MAGN-L2-HIRES_G17 (2018-08-01 to 2023-03-14)"], time0, time1
        )
        result = cdaweb_obj.cda_download(urllist, output_dir)
        self.assertEqual(len(result) > 0, True)
        self.assertTrue(data_exists("b_gse"))

    def test_load_time_clip(self):
        del_data("*")
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(["THEMIS"], ["Magnetic Fields (space)"])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if "THA_L2_FIT" in item), None)

        self.assertTrue(dataset is not None)

        start_time = "2010-01-01 00:00:00"
        end_time = "2010-01-01 12:00:00"

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        cdaweb_obj.cda_download(
            urllist, output_dir, trange=[start_time, end_time], time_clip=True
        )

        # Verify that the times were clipped
        # Note that tplot variables with no data in the requested interval will NOT be clipped (deleted).
        dat = get_data("tha_fgs_dsl")
        self.assertTrue(dat.times[0] >= time_double(start_time))
        self.assertTrue(dat.times[-1] <= time_double(end_time))

    def test_load_time_clip_no_trange(self):
        del_data("*")
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(["THEMIS"], ["Magnetic Fields (space)"])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if "THA_L2_FIT" in item), None)

        self.assertTrue(dataset is not None)

        start_time = "2010-01-01 00:00:00"
        end_time = "2010-01-01 12:00:00"

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level="WARNING") as logs:
            cdaweb_obj.cda_download(urllist, output_dir, time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "No trange specified" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_empty_trange(self):
        del_data("*")
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(["THEMIS"], ["Magnetic Fields (space)"])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if "THA_L2_FIT" in item), None)

        self.assertTrue(dataset is not None)

        start_time = "2010-01-01 00:00:00"
        end_time = "2010-01-01 12:00:00"

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level="WARNING") as logs:
            cdaweb_obj.cda_download(
                urllist, output_dir, trange=[start_time, start_time], time_clip=True
            )
        got_trange_warning = False
        for o in logs.output:
            if "equal" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_backward_trange(self):
        del_data("*")
        cdaweb_obj = CDAWeb()

        ds_list = cdaweb_obj.get_datasets(["THEMIS"], ["Magnetic Fields (space)"])

        # End date of THEMIS mag data is continuously updated, so we have to search the available datesets
        # rather than hardcoding it

        dataset = next((item for item in ds_list if "THA_L2_FIT" in item), None)

        self.assertTrue(dataset is not None)

        start_time = "2010-01-01 00:00:00"
        end_time = "2010-01-01 12:00:00"

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level="WARNING") as logs:
            cdaweb_obj.cda_download(
                urllist, output_dir, trange=[end_time, start_time], time_clip=True
            )
        got_trange_warning = False
        for o in logs.output:
            if "out of order" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_data_prefix(self):
        del_data("*")
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(
            ["VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)"], "1979-01-11", "1979-01-12"
        )
        cdaweb_obj.cda_download(urllist, output_dir, prefix="v2_")
        self.assertTrue(data_exists("v2_B1"))

    def test_load_icon_netcdf_alt_urls(self):
        del_data("*")
        # Create the CDAWeb interface object
        import pyspedas
        from pyspedas.cdagui_tools.config import CONFIG

        # Save the URL settings from the current configuration
        save_cdas_endpoint = CONFIG["cdas_endpoint"]
        save_remote_data_dir = CONFIG["remote_data_dir"]

        # Replace the standard URLs with http rather than https and see if it still works
        alternate_endpoint = "http://cdaweb.gsfc.nasa.gov/WS/cdasr/1/dataviews/sp_phys/"
        alternate_download_url = "http://cdaweb.gsfc.nasa.gov/sp_phys/data"

        CONFIG["cdas_endpoint"] = alternate_endpoint
        CONFIG["remote_data_dir"] = alternate_download_url

        cdaweb_obj = pyspedas.CDAWeb()

        # I'm not sure if it's possible to get the cdasws library to issue a log message with the endpoint URL that
        # was actually used.  So we'll go ahead and peek at the internals, just to see if the object got initialized as expected.
        endpoint_used = cdaweb_obj.cdas._endpoint

        # Note that the time range is different from the other ICON test, to avoid tests interfering with each other
        time0 = "2021-02-15T14:05:52"
        time1 = "2021-02-15T15:08:57"

        urllist = cdaweb_obj.get_filenames(
            ["ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)"],
            time0,
            time1,
        )
        result = cdaweb_obj.cda_download(urllist, output_dir)
        self.assertEqual(len(result) > 0, True)

        # Restore prior configured URLs
        CONFIG["cdas_endpoint"] = save_cdas_endpoint
        CONFIG["remote_data_dir"] = save_remote_data_dir

        # Check that the expected endpoint was used
        self.assertEqual(endpoint_used, alternate_endpoint)
        # Check that the expected download URL was used
        self.assertTrue(alternate_download_url in urllist[0])
        # Check that data was successfully loaded
        self.assertTrue(data_exists("ICON_L22_Fringe_Amplitude"))


if __name__ == "__main__":
    unittest.main()




from pyspedas.projects.themis.load import load


def scm(trange=['2007-03-23', '2007-03-24'],
        probe='c',
        level='l2',
        suffix='',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads Search-coil magnetometer (SCM) data

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2007-03-23', '2007-03-24']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')
            Default: 'c'

        level: str
            Data level; Valid options: 'l1', 'l2'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  
            Default: False; only loads data with a "VAR_TYPE" attribute of "data"

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None; all variables are loaded

        varnames: list of str
            List of variable names to load
            Default: Empty list, so all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword
            Default: False

    Returns
    -------
    List of str
        List of tplot variables created
        Empty list if no data

    Example
    -------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> scm_vars = pyspedas.projects.themis.scm(probe='d', trange=['2013-11-05', '2013-11-06'])
        >>> tplot(['thd_scf_btotal', 'thd_scf_gse'])

    """
    return load(instrument='scm', trange=trange, level=level,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames,
                downloadonly=downloadonly, notplot=notplot,
                probe=probe, time_clip=time_clip, no_update=no_update)



from .load import load


def epd(trange=['2020-06-14', '2020-06-15'],
        datatype='step', 
        mode='hcad', 
        level='l2', 
        suffix='', 
        prefix='', 
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic Particle Detector (EPD)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        mode: str
            EPD mode; Valid options:

        level: str
            Data level (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no prefix is added.
            Default: ''
        
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epd', trange=trange, level=level, datatype=datatype, mode=mode, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)





from pyspedas import find_datasets

# This routine was originally in poes/__init__.py.
def datasets(instrument=None, label=True):
    return find_datasets(mission='POES', instrument='sem2', label=label)




import logging
import numpy as np
from pyspedas import clip, get_data, options, store_data, ylim, zlim

from pyspedas.tplot_tools import time_double


from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional, Iterable, Any

# LEPE data is returned as a dictionary, not tplot variables.  This routine searches the dictionary keys, and renames
# anything that doesn't follow the version v04_01 conventions that the rest of the codebase expects.

def fix_lepe_varnames(var_dict: dict):
    rename_dict = {}
    for varname in var_dict.keys():
        if 'fedo' in varname:
            newname = varname.replace('fedo', 'FEDO')
            rename_dict[varname] = newname
        elif 'fedu' in varname:
            newname = varname.replace('fedu', 'FEDU')
            rename_dict[varname] = newname
        elif 'count_rate_bg' in varname:
            newname = varname.replace('count_rate_bg', 'Count_Rate_BG')
            rename_dict[varname] = newname
        elif 'count_rate' in varname:
            newname = varname.replace('count_rate', 'Count_Rate')
            rename_dict[varname] = newname

    for oldname in rename_dict.keys():
        value = var_dict.pop(oldname)
        newname = rename_dict[oldname]
        logging.info(f'Renaming {oldname} to {newname}')
        var_dict[newname] = value


def lepe(
    trange: List[str] = ['2017-04-04', '2017-04-05'],
    datatype: str = 'omniflux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    version: Optional[str] = None,
    only_fedu: bool = False,
    et_diagram: bool = False,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the LEP-e experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-04-04','2017-04-05']

        datatype: str
            Data type; Valid 'l1' options: None
            Valid 'l2' options: 'omniflux', '3dflux', '3dflux_finech'
            Valid 'l3' options: 'pa'
            Default: 'omniflux'

        level: str
            Data level; Valid options: 'l1','l2','l3'   Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables are loaded)

        varnames: list of str
            List of variable names to load
            Default: [] (all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

        no_update: bool
            If set, only load data from your local cache. Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

        ror: bool
            If set, print PI info and rules of the road. Default: True

        version: str
            Set this value to specify the version of cdf files (such as "v02_02")
            Default: None

        only_fedu: bool
            If set, not make erg_lepe_l3_pa_enech ??(??:01,01,..32)_FEDU Tplot Variables
            Default: False

        et_diagram: bool
            If set, make erg_lepe_l3_pa_pabin ??(??:01,01,..16)_FEDU Tplot Variables
            Default: False

        uname: str
            User name.  Default: None

        passwd: str
            Password.  Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> lepe_vars = pyspedas.projects.erg.lepe(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_lepe_l2_omniflux_FEDO')


    """

    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if level == 'l3':
        datatype = 'pa'

    if ((level == 'l2') and (datatype == 'omniflux')) or \
        ((level == 'l2') and (datatype == '3dflux')) or \
            (level == 'l3'):
        # to avoid failure of creation plot variables (at store_data.py) of lepe
        notplot = True

    file_res = 3600. * 24
    prefix = 'erg_lepe_'+level+'_' + datatype + '_'
    pathformat = 'satellite/erg/lepe/'+level+'/'+datatype + \
        '/%Y/%m/erg_lepe_'+level+'_'+datatype+'_%Y%m%d_'

    if version is None:
        pathformat += 'v??_??.cdf'
    else:
        pathformat += version + '.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)


    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG LEPe')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ",gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            if level == 'l2':
                print(
                    'RoR of LEPe L2: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Lepe')
            if level == 'l3':
                print(
                    'RoR of LEPe L3: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Lepe')
                print(
                    'RoR of MGF L2: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mgf')
            print('')
            print('Contact: erg_lepe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if (isinstance(loaded_data, dict) and (len(loaded_data) > 0)):
        fix_lepe_varnames(loaded_data)

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if (isinstance(loaded_data, dict)) and (len(loaded_data) > 0):
        if (level == 'l2') and (datatype == 'omniflux'):
            tplot_variables = []
            v_keyname = 'v'
            if v_keyname not in loaded_data[prefix + 'FEDO' + suffix]:
                v_keyname = 'v1'
            raw_v1_array = loaded_data[prefix + 'FEDO' + suffix][v_keyname]
            if len(raw_v1_array.shape) == 3:
                # Prior to version 5, the v array included upper and lower bounds that neeeded to be averaged
                v_array = (loaded_data[prefix + 'FEDO' + suffix][v_keyname][:, 0, :] +
                           loaded_data[prefix + 'FEDO' + suffix][v_keyname][:, 1, :]) / 2.
            else:
                # Version 5 doesn't have the extra dimension
                v_array = raw_v1_array
            # change minus values to NaN
            v_array = np.where(v_array < 0., np.nan, v_array)

            if len(v_array.shape) == 2:
                # Version 4 had time-varying v values.  Delete times where v values were all-NaN

                all_nan_v_indices_array = np.where(
                    np.all(np.isnan(v_array), axis=1))[0]
                store_data(prefix + 'FEDO' + suffix,
                           data={'x': np.delete(loaded_data[prefix + 'FEDO' + suffix]['x'], all_nan_v_indices_array, axis=0),
                                 'y': np.delete(loaded_data[prefix + 'FEDO' + suffix]['y'], all_nan_v_indices_array, axis=0),
                                 'v': np.delete(v_array, all_nan_v_indices_array, 0)},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDO' + suffix]['CDF']})
            else:
                # Version 5 has same v array for all times, don't bother cleaning all-NaNs
                store_data(prefix + 'FEDO' + suffix,
                           data={'x': loaded_data[prefix + 'FEDO' + suffix]['x'],
                                 'y': loaded_data[prefix + 'FEDO' + suffix]['y'],
                                 'v': v_array},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDO' + suffix]['CDF']})

            tplot_variables.append(prefix + 'FEDO' + suffix)

            # set spectrogram plot option
            options(prefix + 'FEDO' + suffix, 'Spec', 1)
            # change minus values to NaN in y array
            clip(prefix + 'FEDO' + suffix, 0.,
                 np.nanmax(loaded_data[prefix + 'FEDO' + suffix]['y']))

            # set y axis to logscale
            options(prefix + 'FEDO' + suffix, 'ylog', 1)

            # set ylim
            ylim(prefix + 'FEDO' + suffix, 19, 21*1e3)

            # set ytitle
            options(prefix + 'FEDO' + suffix, 'ytitle',
                    'ERG\nLEP-e\nFEDO\nEnergy')

            # set ysubtitle
            options(prefix + 'FEDO' + suffix, 'ysubtitle', '[eV]')

            # set z axis to logscale
            options(prefix + 'FEDO' + suffix, 'zlog', 1)

            # set zlim
            zlim(prefix + 'FEDO' + suffix,  1, 1e6)

            # set ztitle
            options(prefix + 'FEDO' + suffix, 'ztitle',  '[/cm^{2}-str-s-eV]')

            # change colormap option
            options(prefix + 'FEDO' + suffix, 'Colormap', 'jet')

            return tplot_variables

        if (level == 'l2') and (datatype == '3dflux'):
            tplot_variables = []
            other_variables_dict = {}
            if prefix + 'FEDU' + suffix in loaded_data:
                make_3dflux_tvar(loaded_data, prefix+'FEDU'+suffix, trange)
                tplot_variables.append(prefix + 'FEDU' + suffix)

            if prefix + 'Count_Rate' + suffix in loaded_data:
                make_3dflux_tvar(loaded_data, prefix+'Count_Rate'+suffix, trange)
                tplot_variables.append(prefix + 'Count_Rate' + suffix)
                other_variables_dict[prefix + 'Count_Rate' +
                                     suffix] = loaded_data[prefix + 'Count_Rate' + suffix]


            if prefix + 'Count_Rate_BG' + suffix in loaded_data:
                # Can't use make_3dflux_tvar here because this variable doesn't have a v3 dimension
                varname=prefix + 'Count_Rate_BG'+suffix
                trange_double = np.array(time_double(trange))
                trange_dt64 = (trange_double * 1.0e6).astype('datetime64[us]')
                time_array = np.array(loaded_data[varname]['x'])
                inside_indices_array = np.argwhere((trange_dt64[0] < time_array)
                                                   & (trange_dt64[1] > time_array))
                inside_indices_list = inside_indices_array[:, 0].tolist()
                v_keyname = 'v'
                if v_keyname not in loaded_data[varname].keys():
                    v_keyname = 'v1'
                raw_v1_array = loaded_data[varname][v_keyname]

                if len(raw_v1_array.shape) == 3:
                    # Prior to version 5, the v1 array had three dimensions: time, upper/lower, energy
                    # The upper/lower bounds need to be averaged to get center values.
                    # For version 4 and earlier, the data dimensions are time, sector, energy.

                    store_data(varname,
                               data={'x': time_array[inside_indices_list],
                                     'y': loaded_data[varname]['y'][inside_indices_list],
                                     'v1': (raw_v1_array[inside_indices_list][:, 0, :]
                                            + raw_v1_array[inside_indices_list][:, 1, :]) / 2.,  # arithmetic mean
                                     'v2': [i for i in range(16)],
                                    },
                               attr_dict={'CDF': loaded_data[varname]['CDF']})
                else:
                    # Version 5 has a one-dimensional v1 array (which is no longer the energy...it's the sector/phase)
                    # Dimensions are now time x 16 x 32 x 12
                    # v1, v2 are present in the metadata, v2 is time-varying energies
                    # For version 4 and earlier, the data dimensions are time, sector, energy.
                    # With version 5, the data dimensions are time, energy, phase/sector
                    # So the last two dimensions and their metadata need to be swapped for V5 and later

                    data_array_filtered = loaded_data[varname]['y'][inside_indices_list]
                    data_array_permuted = np.moveaxis(data_array_filtered, (1, 2), (2,1))
                    store_data(varname,
                               data={'x': time_array[inside_indices_list],
                                     'y': data_array_permuted,
                                     'v1': loaded_data[varname]['v2'][inside_indices_list],
                                     'v2': raw_v1_array,
                                    },
                               attr_dict={'CDF': loaded_data[varname]['CDF']})

                options(varname, 'spec', 1)
                ylim(varname, 19, 21 * 1e3)
                zlim(varname, 1, 1e6)
                options(varname, 'zlog', 1)
                options(varname, 'ylog', 1)
                options(varname, 'ysubtitle', '[eV]')

                tplot_variables.append(prefix + 'Count_Rate_BG' + suffix)
                other_variables_dict[prefix + 'Count_Rate_BG' +
                                     suffix] = loaded_data[prefix + 'Count_Rate_BG' + suffix]

                return tplot_variables

        if level == 'l3':
            tplot_variables = []

            if prefix + 'FEDU' + suffix in loaded_data:
                raw_v1_array = loaded_data[prefix + 'FEDU' + suffix]['v1']
                if len(raw_v1_array.shape) == 3:
                    # Prior to version 5, v1 array had dimensions time, upper/lower, energy
                    # The upper/lower values need to be averaged to get the bin centers
                    store_data(prefix + 'FEDU' + suffix,
                               data={'x': loaded_data[prefix + 'FEDU' + suffix]['x'],
                                     'y': loaded_data[prefix + 'FEDU' + suffix]['y'],
                                     'v1': (raw_v1_array[:, 0, :]
                                            + raw_v1_array[:, 1, :]) / 2.,  # arithmetic mean
                                     'v2': loaded_data[prefix + 'FEDU' + suffix]['v2']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU' + suffix]['CDF']})
                else:
                    # In version 5, the v1 array doesn't have the extra dimension
                    store_data(prefix + 'FEDU' + suffix,
                               data={'x': loaded_data[prefix + 'FEDU' + suffix]['x'],
                                     'y': loaded_data[prefix + 'FEDU' + suffix]['y'],
                                     'v1': raw_v1_array,
                                     'v2': loaded_data[prefix + 'FEDU' + suffix]['v2']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU' + suffix]['CDF']})

                tplot_variables.append(prefix + 'FEDU' + suffix)

                options(prefix + 'FEDU' + suffix, 'spec', 1)
                if prefix + 'FEDU' + suffix in tplot_variables:
                    clip(prefix + 'FEDU' + suffix, 0,
                         np.nanmax(loaded_data[prefix + 'FEDU' + suffix]['y']))
                ylim(prefix + 'FEDU' + suffix, 19, 21*1e3)
                zlim(prefix + 'FEDU' + suffix, 1, 1e6)
                options(prefix + 'FEDU' + suffix, 'zlog', 1)
                options(prefix + 'FEDU' + suffix, 'ylog', 1)
                options(prefix + 'FEDU' + suffix, 'ysubtitle', '[eV]')

                FEDU_get_data = get_data(prefix + 'FEDU' + suffix)
                FEDU_CDF_data = loaded_data[prefix + 'FEDU' + suffix]['CDF']

                if not only_fedu:
                    # version 5: 32-element 1-D array
                    # version 4: times x 32, takes the first sample and fixes nans
                    ev_array = FEDU_get_data[2]
                    if len(ev_array.shape) == 2:
                        # version 4: times x 32, take first sample and fix nans
                        ytitle_eV_array = np.round(np.nan_to_num(ev_array[0, :]), 2)
                    else:
                        # version 5: fix all nans
                        ytitle_eV_array = np.round(np.nan_to_num(ev_array[:]), 2)

                    # processing for erg_lepe_l3_pa_enech_??(??:01,01,..32)_FEDU
                    for i in range(FEDU_get_data[1].shape[1]):
                        tplot_name = prefix + 'enech_' + \
                            str(i + 1).zfill(2) + '_FEDU' + suffix
                        store_data(tplot_name, data={'x': FEDU_get_data[0],
                                                     'y': FEDU_get_data[1][:, i, :],
                                                     'v': FEDU_get_data[3]},
                                    attr_dict={'CDF':FEDU_CDF_data})
                        options(tplot_name, 'spec', 1)
                        ylim(tplot_name, 0, 180)
                        zlim(tplot_name, 1, 1e6)
                        options(tplot_name, 'ytitle', 'ERG LEP-e\n' +
                                str(ytitle_eV_array[i]) + ' eV\nPitch angle')
                        tplot_variables.append(tplot_name)

                    options(tplot_variables[1:], 'zlog', 1)
                    options(tplot_variables[1:], 'ysubtitle', '[deg]')
                    options(tplot_variables[1:], 'yrange', [0, 180])
                    options(tplot_variables[1:], 'colormap', 'jet')
                    options(tplot_variables[1:],
                            'ztitle', '[/s-cm^{2}-sr-keV/q]')

                if et_diagram:
                    ytitle_deg_array = np.round(
                        np.nan_to_num(FEDU_get_data[3]), 3)
                    # Version 5: ev_array is time-varying, remove times with all-NaN#
                    # Version 4: constant ev_array, don't delete anything
                    ev_array = FEDU_get_data[2]
                    if len(ev_array.shape) == 2:
                        all_nan_v_indices_array = np.where(
                            np.all(np.isnan(ev_array), axis=1))[0]
                    else:
                        all_nan_v_indices_array = []
                    x_all_nan_deleted_array = np.delete(
                        FEDU_get_data[0], all_nan_v_indices_array, axis=0)
                    y_all_nan_deleted_array = np.delete(
                        FEDU_get_data[1], all_nan_v_indices_array, axis=0)
                    v_all_nan_deleted_array = np.delete(
                        FEDU_get_data[2], all_nan_v_indices_array, axis=0)
                    # processing for erg_lepe_l3_pa_pabin_??(??:01,01,..16)_FEDU
                    for i in range(FEDU_get_data[1].shape[2]):
                        tplot_name = prefix + 'pabin_' + \
                            str(i + 1).zfill(2) + '_FEDU' + suffix
                        store_data(tplot_name, data={'x': x_all_nan_deleted_array,
                                                     'y': y_all_nan_deleted_array[:, :, i],
                                                     'v': v_all_nan_deleted_array},
                                    attr_dict={'CDF':FEDU_CDF_data})
                        options(tplot_name, 'spec', 1)
                        ylim(tplot_name,  19, 21*1e3)
                        zlim(tplot_name, 1, 1e6)
                        options(tplot_name, 'ytitle', 'ERG LEP-e\n' +
                                str(ytitle_deg_array[i]) + ' deg\nEnergy')
                        tplot_variables.append(tplot_name)

                    options(
                        tplot_variables[-FEDU_get_data[1].shape[2]:], 'ysubtitle', '[eV]')
                    options(
                        tplot_variables[-FEDU_get_data[1].shape[2]:], 'zlog', 1)
                    options(
                        tplot_variables[-FEDU_get_data[1].shape[2]:], 'ylog', 1)
                    options(
                        tplot_variables[-FEDU_get_data[1].shape[2]:], 'colormap', 'jet')
                    options(
                        tplot_variables[-FEDU_get_data[1].shape[2]:], 'ztitle', '[/s-cm^{2}-sr-eV]')

                return tplot_variables

    return loaded_data


def make_3dflux_tvar(loaded_data: dict[Any, Any],varname: str, trange: Iterable[str]):
    trange_double = np.array(time_double(trange))
    trange_dt64 = (trange_double * 1.0e6).astype('datetime64[us]')
    time_array = np.array(loaded_data[varname]['x'])
    inside_indices_array = np.argwhere((trange_dt64[0] < time_array)
                                       & (trange_dt64[1] > time_array))
    inside_indices_list = inside_indices_array[:, 0].tolist()
    v_keyname = 'v'
    if v_keyname not in loaded_data[varname].keys():
        v_keyname = 'v1'
    raw_v1_array = loaded_data[varname][v_keyname]

    if len(raw_v1_array.shape) == 3:
        # Prior to version 5, the v1 array had three dimensions: time, upper/lower, energy
        # The upper/lower bounds need to be averaged to get center values.
        # Data array dimensions are time x 32 x 12 x 16  (time, energy, channel, phase)
        store_data(varname,
                   data={'x': time_array[inside_indices_list],
                         'y': loaded_data[varname]['y'][inside_indices_list],
                         'v1': (raw_v1_array[inside_indices_list][:, 0, :]
                                + raw_v1_array[inside_indices_list][:, 1, :]) / 2.,  # arithmetic mean
                         'v2': ['01', '02', '03', '04', '05', 'A', 'B', '18', '19', '20', '21', '22'],
                         'v3': [i for i in range(16)]},
                   attr_dict={'CDF': loaded_data[varname]['CDF']})
    else:
        # Version 5 has a one-dimensional v1 array (which is now the sector, not the energy!)
        # Dimensions are now time x 16 x 32 x 12 (time, sector, energy, channel)
        # v1, v2, v3 are all present in the metadata, v2 is time-varying energies
        # Note that the axis order has changed from version 4!
        # Version 4: time, energy, channel, phase
        # Version5: time, phase, energy, channel
        # So we need to permute the axes and the metadata to ensure that lepe_get_dist will still work.
        data_array_filtered = loaded_data[varname]['y'][inside_indices_list]
        data_array_permuted = np.moveaxis(data_array_filtered, (1,2,3), (3,1,2))
        store_data(varname,
                   data={'x': time_array[inside_indices_list],
                         'y':data_array_permuted,
                         'v1': loaded_data[varname]['v2'][inside_indices_list], # energies
                         'v2': loaded_data[varname]['v3'], # channels (string valued!)
                          # The v1 in the CDF is 1-based, but we need it to be zero-based
                         'v3': [i for i in range(16)] # sectors
                         },
                   attr_dict={'CDF': loaded_data[varname]['CDF']})


    options(varname, 'spec', 1)
    ylim(varname, 19, 21 * 1e3)
    zlim(varname, 1, 1e6)
    options(varname, 'zlog', 1)
    options(varname, 'ylog', 1)
    options(varname, 'ysubtitle', '[eV]')



import numpy as np


def slice2d_orientslice(vectors=None, vbulk=None, bfield=None, sunvec=None, slice_x=None, slice_z=None):
    """
    Performs transformation into user specified coordinates.
    This transformation is applied after the CUSTOM_ROTATION and ROTATION transformation have been performed.
    """
    if slice_z is None:
        slice_z = np.array([0, 0, 1])

    # normalize
    z = slice_z/np.linalg.norm(slice_z)

    # if the x-axis direction is not specified, use projection
    # of the ordinate closest to the slice plane
    if slice_x is None:
        ordinate = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        angles = (180.0/np.pi)*np.arccos(np.array([np.sum(ordinate[:, 0]*z),
                                                   np.sum(ordinate[:, 1]*z),
                                                   np.sum(ordinate[:, 2]*z)]))

        # use ordinate that is closest to perpendicular with norm
        ind = np.argmin(np.abs(angles-90.0))

        slice_x = ordinate[:, ind]

    # normalize vector used to define slice's x-axis
    xp = slice_x/np.linalg.norm(slice_x)

    # get the slice's y-axis
    y = np.cross(z, xp)

    # get the slice's x-axis
    x = np.cross(y, z)

    # now normalize the x-axis
    x = x/np.linalg.norm(x)

    # get the rotation matrix
    matrix = np.array([x, y, z])

    # Transform particle and support vectors
    if vectors is not None:
        transformed = np.zeros(vectors.shape)
        num_vectors = vectors.shape[0]
        for vector_idx in range(num_vectors):
            transformed[vector_idx] = (matrix.T @ vectors[vector_idx, :]).flatten()
        vectors = transformed
    if vbulk is not None:
        vbulk = matrix @ vbulk
    if bfield is not None:
        bfield = matrix @ bfield
    if sunvec is not None:
        sunvec = matrix @ sunvec

    return {'matrix': matrix,
            'vectors': vectors,
            'vbulk': vbulk,
            'bfield': bfield,
            'sunvec': sunvec}




import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data, timespan, tplot, options
from pyspedas.projects.erg import erg_mep_part_products

import pyspedas
import pyspedas

display=False

class LoadTestCases(unittest.TestCase):

    def test_mepe_theta(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', outputs='theta', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot( 'erg_mepe_l2_3dflux_FEDU_theta', display=display, save_png='erg_mepe_theta.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_theta' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_theta'))

    def test_mepe_theta_no_trange(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', outputs='theta' )
        tplot( 'erg_mepe_l2_3dflux_FEDU_theta', display=display, save_png='erg_mepe_theta_no_trange.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_theta' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_theta'))

    def test_mepe_phi(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', outputs='phi', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot( 'erg_mepe_l2_3dflux_FEDU_phi', display=display, save_png='erg_mepe_phi.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_phi' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_phi'))


    def test_mepe_pa(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', mag_name=mag_vn, pos_name=pos_vn, outputs='pa', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pa.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_gyro(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', mag_name=mag_vn, pos_name=pos_vn, outputs='gyro', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot( 'erg_mepe_l2_3dflux_FEDU_gyro', display=display, save_png='erg_mepe_gyro.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_gyro' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_gyro'))

    def test_mepe_moments(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', mag_name=mag_vn, pos_name=pos_vn, outputs='moments', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot(vars, display=display, save_png='erg_mepe_moments.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_density' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_density'))
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_density' in vars)
        self.assertEqual(pyspedas.get_units('erg_mepe_l2_3dflux_FEDU_density'), '1/cm^3')
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_qflux'))
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_qflux' in vars)
        self.assertEqual(pyspedas.get_units('erg_mepe_l2_3dflux_FEDU_qflux'), 'eV/(cm^2-sec)')
        self.assertEqual(pyspedas.get_coords('erg_mepe_l2_3dflux_FEDU_qflux').lower(), 'dsi')


    def test_mepe_fac_moments(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', mag_name=mag_vn, pos_name=pos_vn, outputs='fac_moments', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot(vars, display=display, save_png='erg_mepe_fac_moments.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_density_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_density_mag'))

    def test_mepe_fac_energy(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', mag_name=mag_vn, pos_name=pos_vn, outputs='fac_energy', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot(vars, display=display, save_png='erg_mepe_fac_energy.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy_mag'))

    def test_mepe_energy(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        # Calculate and plot energy spectrum
        vars = erg_mep_part_products( 'erg_mepe_l2_3dflux_FEDU', outputs='energy', trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'] )
        tplot( 'erg_mepe_l2_3dflux_FEDU_energy', display=display, save_png='erg_mepe_en_spec.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy'))


    def test_mepe_pad_xdsi(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.], fac_type='xdsi',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pad_xdsi.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_pad_mphigeo(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.], fac_type='mphigeo',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pad_mphigeo.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_pad_phigeo(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.], fac_type='phigeo',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_pad_xgse(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux')
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.],
                                     fac_type='xgse',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pad_xgse.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_pad_phism(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux')
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.],
                                     fac_type='phism',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pad_phism.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))

    def test_mepe_pad_mphism(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux')
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate the pitch angle distribution
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='pa', energy=[15000., 22000.],
                                     fac_type='mphism',
                                     mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        tplot( 'erg_mepe_l2_3dflux_FEDU_pa', display=display, save_png='erg_mepe_pad_mphism.png' )
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_pa' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_pa'))


    def test_mepe_en_pad_limit_gyro(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate energy-time spectra of electron flux for limited pitch-angle (PA) ranges
        ## Here we calculate energy-time spectra for PA = 0-10 deg and PA = 80-100 deg.
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='fac_energy', gyro=[0.0, 90.0],
                                     fac_type='xdsi', mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        ## Decorate the obtained spectrum variables
        options('erg_mepe_l2_3dflux_FEDU_energy_mag', 'ytitle', 'MEP-e flux\nPA: 80-100\n\n[eV]')
        tplot(['erg_mepe_l2_3dflux_FEDU_energy_mag'], display=display, save_png='erg_mep_en_pa_limit_gyro.png')
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy_mag'))


    def test_mepe_en_pad_limit_theta(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate energy-time spectra of electron flux for limited pitch-angle (PA) ranges
        ## Here we calculate energy-time spectra for PA = 0-10 deg and PA = 80-100 deg.
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='fac_energy', theta=[0.0, 90.0],
                                     fac_type='xdsi', mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        ## Decorate the obtained spectrum variables
        options('erg_mepe_l2_3dflux_FEDU_energy_mag', 'ytitle', 'MEP-e flux\nPA: 80-100\n\n[eV]')
        tplot(['erg_mepe_l2_3dflux_FEDU_energy_mag'], display=display, save_png='erg_mep_en_pa_limit_theta.png')
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy_mag'))

    def test_mepe_en_pad_limit_phi_in(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate energy-time spectra of electron flux for limited pitch-angle (PA) ranges
        ## Here we calculate energy-time spectra for PA = 0-10 deg and PA = 80-100 deg.
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='fac_energy', phi_in=[0.0, 180.0],
                                     fac_type='xdsi', mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        ## Decorate the obtained spectrum variables
        options('erg_mepe_l2_3dflux_FEDU_energy_mag', 'ytitle', 'MEP-e flux\nPA: 80-100\n\n[eV]')
        tplot(['erg_mepe_l2_3dflux_FEDU_energy_mag'], display=display, save_png='erg_mep_en_pa_limit_phi_in.png')
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy_mag'))

    def test_mepe_en_pad_limit_energy(self):
        del_data('*')
        # Load MEP-e Lv.2 3-D flux data
        timespan('2017-04-05 21:45:00', 2.25, keyword='hours')
        pyspedas.projects.erg.mepe( trange=[ '2017-04-05 21:45:00', '2017-04-05 23:59:59'], datatype='3dflux' )
        vars = pyspedas.projects.erg.mgf(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary B-field data
        vars = pyspedas.projects.erg.orb(trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])  # Load necessary orbit data
        mag_vn = 'erg_mgf_l2_mag_8sec_dsi'
        pos_vn = 'erg_orb_l2_pos_gse'
        # Calculate energy-time spectra of electron flux for limited pitch-angle (PA) ranges
        ## Here we calculate energy-time spectra for PA = 0-10 deg and PA = 80-100 deg.
        vars = erg_mep_part_products('erg_mepe_l2_3dflux_FEDU', outputs='fac_energy', energy=[15000., 22000.],
                                     fac_type='xdsi', mag_name=mag_vn, pos_name=pos_vn,
                                     trange=['2017-04-05 21:45:00', '2017-04-05 23:59:59'])
        ## Decorate the obtained spectrum variables
        options('erg_mepe_l2_3dflux_FEDU_energy_mag', 'ytitle', 'MEP-e flux\nPA: 80-100\n\n[eV]')
        tplot(['erg_mepe_l2_3dflux_FEDU_energy_mag'], display=display, save_png='erg_mep_en_pa_limit_energy.png')
        self.assertTrue('erg_mepe_l2_3dflux_FEDU_energy_mag' in vars)
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU_energy_mag'))

if __name__ == '__main__':
    unittest.main()


from functools import update_wrapper
from pyspedas.utilities.pyspedas_functools import better_partial
from .load import load
from pyspedas.utilities.datasets import find_datasets


# Define partial wrappers for other load routines, fixing the instrument parameter
# 'better_partial' works better with PyCharm autocompletion than functools.partial
# update_wrapper() is necessary for help() to show info for the wrapped function,
# rather than the partial() object.

dcf = better_partial(load, instrument="dcf")
update_wrapper(dcf, load)
dcf.__doc__ = """
    Load FAST dcf data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1996-12-01", "1996-12-02"].
    instrument : str, optional
        Type of instrument.
        Values can be: 'dcf', 'acf', 'esa', 'teams'.

        .. note::
           For this function, the `instrument` parameter is preset to `'dcf'` and should not be set manually.

    datatype : str, optional
        Data type to load. Depends on the instrument.
        For 'esa' valid options are: 'eeb', 'ees', 'ieb', 'ies'.
        For all other insturments, this keyword is ignored.
        Default is ''.
    level : str, optional
        Data level to load. Depends on the instrument.
        For 'dcf' and 'teams' valid options are: 'l2', 'k0'.
        For all other instruments, this keyword is ignored.
        Default is 'l2'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    """


acf = better_partial(load, instrument="acf")
update_wrapper(acf, load)
acf.__doc__ = """
    Load FAST acf data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1996-12-01", "1996-12-02"].
    instrument : str, optional
        Type of instrument.
        Values can be: 'dcf', 'acf', 'esa', 'teams'.

        .. note::
           For this function, the `instrument` parameter is preset to `'acf'` and should not be set manually.

    datatype : str, optional
        Data type to load. Depends on the instrument.
        For 'esa' valid options are: 'eeb', 'ees', 'ieb', 'ies'.
        For all other insturments, this keyword is ignored.
        Default is ''.
    level : str, optional
        Data level to load. Depends on the instrument.
        For 'dcf' and 'teams' valid options are: 'l2', 'k0'.
        For all other instruments, this keyword is ignored.
        Default is 'l2'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    """


esa = better_partial(load, instrument="esa")
update_wrapper(esa, load)
esa.__doc__ = """
    Load FAST esa data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1996-12-01", "1996-12-02"].
    instrument : str, optional
        Type of instrument.
        Values can be: 'dcf', 'acf', 'esa', 'teams'.

        .. note::
           For this function, the `instrument` parameter is preset to `'esa'` and should not be set manually.

    datatype : str, optional
        Data type to load. Depends on the instrument.
        For 'esa' valid options are: 'eeb', 'ees', 'ieb', 'ies'.
        For all other insturments, this keyword is ignored.
        Default is ''.
    level : str, optional
        Data level to load. Depends on the instrument.
        For 'dcf' and 'teams' valid options are: 'l2', 'k0'.
        For all other instruments, this keyword is ignored.
        Default is 'l2'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    """


teams = better_partial(load, instrument="teams")
update_wrapper(teams, load)
teams.__doc__ = """
    Load FAST teams data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1996-12-01", "1996-12-02"].
    instrument : str, optional
        Type of instrument.
        Values can be: 'dcf', 'acf', 'esa', 'teams'.

        .. note::
           For this function, the `instrument` parameter is preset to `'teams'` and should not be set manually.

    datatype : str, optional
        Data type to load. Depends on the instrument.
        For 'esa' valid options are: 'eeb', 'ees', 'ieb', 'ies'.
        For all other insturments, this keyword is ignored.
        Default is ''.
    level : str, optional
        Data level to load. Depends on the instrument.
        For 'dcf' and 'teams' valid options are: 'l2', 'k0'.
        For all other instruments, this keyword is ignored.
        Default is 'l2'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    """


datasets = better_partial(find_datasets, mission="FAST", label=True)
update_wrapper(datasets, find_datasets)



# Copyright 2020 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pyspedas
from pyspedas.tplot_tools import store_data, tinterp
import copy
import logging

def subtract(
        tvar1,
        tvar2,
        newname=None,
):
    """
    Subtracts two tplot variables.  Will interpolate if the two are not on the same time cadence.

    Parameters
    ----------
        tvar1 : str
            Name of first tplot variable.
        tvar2 : int/float
            Name of second tplot variable
        newname : str
            Name of new tvar for added data.
            Default: None. If not set, then the data in tvar1 is replaced.

    Returns
    -------
        None

    Examples
    --------
        >>> pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pyspedas.store_data('c', data={'x':[0,4,8,12,16,19,21], 'y':[1,4,1,7,1,9,1]})
        >>> pyspedas.subtract('a','c',newname='a-c')

    """

    #interpolate tvars
    tv2 = tinterp(tvar1,tvar2)

    #separate and subtract data
    data1 = pyspedas.tplot_tools.data_quants[tvar1].values
    data2 = pyspedas.tplot_tools.data_quants[tv2].values
    data = data1 - data2

    #store subtracted data
    if newname is None:
        pyspedas.tplot_tools.data_quants[tvar1].values = data
        return tvar1

    if 'spec_bins' in pyspedas.tplot_tools.data_quants[tvar1].coords:
        store_data(newname, data={'x': pyspedas.tplot_tools.data_quants[tvar1].coords['time'].values, 'y': data, 'v':pyspedas.tplot_tools.data_quants[tvar1].coords['spec_bins'].values})
        pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar1].attrs)
    else:
        store_data(newname,data={'x': pyspedas.tplot_tools.data_quants[tvar1].coords['time'].values, 'y': data})
        pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar1].attrs)

    return newname



import logging
import numpy as np

def replace_fillvals(data_array, fillval, varname, param_type):
    """
    Replace fill values in data_array with NaNs.  If fillval is an array, each entry applies to the correspondiing data_array elements.

    Parameters
    ----------
    data_array: ndarray
        Input data
    fillval: int or float or list or ndarray
        Fill values to replace with NaN in corresponding elements of data_array.
    varname: str
        The variable currently being processed (for log messages)
    param_type: str
        Data type as returned by hapi server.  Integer data is 0-filled, while floating point data is nan-filled.

    Returns
    -------
    None

    Notes
    -----
    If fillval is an array, its shape must match that of the data array values (after removing the leading dimension for
    the timestamps.  Each element of fillval is searched for and replaced independently, in the correspondina array positions of the data array.

    """

    if param_type == 'integer':
        replacement = 0
    else:
        replacement = np.nan

    if np.isscalar(fillval):
        # Scalar fillval, replace scalar fillval everywhere
        idx = np.where(data_array == fillval)
        data_array[idx] = replacement
    else:
        logging.warning(f'replace_fillvals: Fillval for {varname} is a array-valued, which may not be standard-compliant.')
        # Array fillval, check that shapes match, and replace data_array values matching corresponding fillvals
        npfillval=np.array(fillval)
        npfillval_shape = npfillval.shape
        data_array_shape = data_array.shape

        if npfillval_shape != data_array[0].shape:
            logging.error(f'replace_fillvals: Fillval shape {str(npfillval_shape)} not compatible with data_array shape {str(data_array_shape)}')
            return
        npfillval_reshaped = npfillval.reshape(-1)
        data_reshaped = data_array.reshape(data_array_shape[0],-1)
        for i in range(len(npfillval_reshaped)):
            idx = np.where(data_reshaped[:,i] == npfillval_reshaped[i])
            data_reshaped[idx,i] = replacement








import numpy as np
import pyspedas

def data_exists(tvar):
    """
    Checks if a tplot variable exists

    Parameters
    ----------
    tvar: str
        Name of tplot variable to check

    Returns
    --------
    bool
        Retrun True if tplot variable exists.

    Example
    ----------
        >>> import pyspedas
        >>> pyspedas.store_data('a', data={'x': range(10), 'y': range(10)})
        >>> pyspedas.data_exists('a')

    """
    if tvar in pyspedas.tplot_tools.data_quants.keys():
        data = pyspedas.get_data(tvar, dt=True)
        # multi-dimensional data returns a tuple, NRV variables return an ndarray
        if isinstance(data, tuple) or isinstance(data, np.ndarray) or isinstance(data, str) or isinstance(data, list):
            return True
    return False



"""
Unit Tests for minvar function.
"""

import os
import unittest
import numpy as np
from numpy.testing import assert_allclose
from pyspedas import (
    fac_matrix_make,
    get_data,
    tplot_copy,
    del_data,
    store_data,
    tplot_restore,
    tvector_rotate,
    rotmat_get_coords,
    tplot,
    get_coords,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
cotrans_dir = "cotrans_tools"
save_dir = os.path.join(output_dir, cotrans_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class TestFac(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        #  Test tolerance
        cls.tol = 1e-10

        # Define a random data array
        rng = np.random.default_rng(seed=31415)
        cls.rdata = rng.random((10, 3))

        # Download tplot files
        # Comparison data was generated using both thm_fac_matrix_make and fac_matrix_make, using the script
        # general/tools/python_validate/fac_matrix_make_python_validate

        # Load validation variables from the test file
        filename = test_data_download_file(
            validation_dir, cotrans_dir, "fac_python_validate.tplot", save_dir
        )

        # Load validation variables from the test file
        del_data("*")
        tplot_restore(filename)

        cls.fgm = get_data("thc_fgs_gse")
        tplot_copy("thc_fgs_gse", "idl_thc_fgs_gse")
        cls.fgm_sm = get_data("thc_fgs_gse_sm601")
        tplot_copy("thc_fgs_gse_sm601", "idl_thc_fgs_gse_sm601")
        cls.pos = get_data("thc_state_pos")
        tplot_copy("thc_state_pos", "idl_thc_state_pos")

        cls.mat_xgse = get_data("thc_fgs_gse_sm601_fac_mat_xgse")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_xgse", "idl_thc_fgs_gse_sm601_fac_mat_xgse"
        )
        cls.mat_rgeo = get_data("thc_fgs_gse_sm601_fac_mat_rgeo")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_rgeo", "idl_thc_fgs_gse_sm601_fac_mat_rgeo"
        )
        cls.mat_phigeo = get_data("thc_fgs_gse_sm601_fac_mat_phigeo")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phigeo", "idl_thc_fgs_gse_sm601_fac_mat_phigeo"
        )
        cls.mat_mphigeo = get_data("thc_fgs_gse_sm601_fac_mat_mphigeo")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phism", "idl_thc_fgs_gse_sm601_fac_mat_phism"
        )
        cls.mat_phism = get_data("thc_fgs_gse_sm601_fac_mat_phism")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phism", "idl_thc_fgs_gse_sm601_fac_mat_phism"
        )
        cls.mat_mphism = get_data("thc_fgs_gse_sm601_fac_mat_mphism")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phism", "idl_thc_fgs_gse_sm601_fac_mat_phism"
        )
        cls.mat_ygsm = get_data("thc_fgs_gse_sm601_fac_mat_ygsm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_ygsm", "idl_thc_fgs_gse_sm601_fac_mat_ygsm"
        )

        cls.mat_xgse_thm = get_data("thc_fgs_gse_sm601_fac_mat_xgse_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_xgse_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_xgse_thm",
        )
        cls.mat_rgeo_thm = get_data("thc_fgs_gse_sm601_fac_mat_rgeo_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_rgeo_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_rgeo_thm",
        )
        cls.mat_phigeo_thm = get_data("thc_fgs_gse_sm601_fac_mat_phigeo_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phigeo_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_phigeo_thm",
        )
        cls.mat_mphigeo_thm = get_data("thc_fgs_gse_sm601_fac_mat_mphigeo_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_mphigeo_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_mphigeo_thm",
        )
        cls.mat_phism_thm = get_data("thc_fgs_gse_sm601_fac_mat_phism_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_phism_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_phism_thm",
        )
        cls.mat_mphism_thm = get_data("thc_fgs_gse_sm601_fac_mat_mphism_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_mphism_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_mphism_thm",
        )
        cls.mat_ygsm_thm = get_data("thc_fgs_gse_sm601_fac_mat_ygsm_thm")
        tplot_copy(
            "thc_fgs_gse_sm601_fac_mat_ygsm_thm",
            "idl_thc_fgs_gse_sm601_fac_mat_ygsm_thm",
        )

        cls.vec_xgse = get_data("thc_fgs_fac_xgse")
        tplot_copy("thc_fgs_fac_xgse", "idl_thc_fgs_fac_xgse")
        cls.vec_rgeo = get_data("thc_fgs_fac_rgeo")
        tplot_copy("thc_fgs_fac_rgeo", "idl_thc_fgs_fac_rgeo")
        cls.vec_phigeo = get_data("thc_fgs_fac_phigeo")
        tplot_copy("thc_fgs_fac_phigeo", "idl_thc_fgs_fac_phigeo")
        cls.vec_mphigeo = get_data("thc_fgs_fac_mphigeo")
        tplot_copy("thc_fgs_fac_mphigeo", "idl_thc_fgs_fac_mphigeo")
        cls.vec_phism = get_data("thc_fgs_fac_phism")
        tplot_copy("thc_fgs_fac_phism", "idl_thc_fgs_fac_phism")
        cls.vec_mphism = get_data("thc_fgs_fac_mphism")
        tplot_copy("thc_fgs_fac_mphism", "idl_thc_fgs_fac_mphism")
        cls.vec_ygsm = get_data("thc_fgs_fac_ygsm")
        tplot_copy("thc_fgs_fac_ygsm", "idl_thc_fgs_fac_ygsm")

        cls.vec_xgse_thm = get_data("thc_fgs_fac_xgse_thm")
        tplot_copy("thc_fgs_fac_xgse_thm", "idl_thc_fgs_fac_xgse_thm")
        cls.vec_rgeo_thm = get_data("thc_fgs_fac_rgeo_thm")
        tplot_copy("thc_fgs_fac_rgeo_thm", "idl_thc_fgs_fac_rgeo_thm")
        cls.vec_phigeo_thm = get_data("thc_fgs_fac_phigeo_thm")
        tplot_copy("thc_fgs_fac_phigeo_thm", "idl_thc_fgs_fac_phigeo_thm")
        cls.vec_mphigeo_thm = get_data("thc_fgs_fac_mphigeo_thm")
        tplot_copy("thc_fgs_fac_mphigeo_thm", "idl_thc_fgs_fac_mphigeo_thm")
        cls.vec_phism_thm = get_data("thc_fgs_fac_phism")
        tplot_copy("thc_fgs_fac_phism_thm", "idl_thc_fgs_fac_phism_thm")
        cls.vec_mphism_thm = get_data("thc_fgs_fac_mphism")
        tplot_copy("thc_fgs_fac_mphism_thm", "idl_thc_fgs_fac_mphism_thm")
        cls.vec_ygsm_thm = get_data("thc_fgs_fac_ygsm_thm")
        tplot_copy("thc_fgs_fac_ygsm_thm", "idl_thc_fgs_fac_ygsm_thm")

        del_data("thc_*")

    def test_fac_xgse(self):
        """Test of other_dim = xgse"""

        fac_matrix_make("idl_thc_fgs_gse_sm601", other_dim="xgse", newname="mat_xgse")
        fac_in, fac_out = rotmat_get_coords("mat_xgse")
        self.assertEqual(fac_in, "GSE")
        self.assertEqual(fac_out, "FAC-XGSE")
        tvector_rotate("mat_xgse", "idl_thc_fgs_gse", newname="vec_xgse")
        out_vec_coords = get_coords("vec_xgse")
        self.assertEqual(out_vec_coords, "FAC-XGSE")
        dat1 = get_data("mat_xgse")
        dat2 = get_data("vec_xgse")
        store_data("vecdiff", data={"x": dat2.times, "y": self.vec_xgse.y - dat2.y})
        assert_allclose(dat1.y, self.mat_xgse.y, atol=1.0e-06)
        assert_allclose(dat2.y, self.vec_xgse.y, atol=1.0e-06)
        store_data("vecdiff", data={"x": dat2.times, "y": self.vec_xgse.y - dat2.y})
        # Test rotmat_get_coords on variable that doesn't have the right metadata
        bogus_in, bogus_out = rotmat_get_coords("vecdiff")
        self.assertEqual(bogus_in, None)
        self.assertEqual(bogus_out, None)
        # Test tvector_rotate on a variable with no coords set
        tvector_rotate("mat_xgse", "vecdiff", newname="bogus_vec_cotrans")
        local_png = os.path.join(save_dir, "fac_xgse.png")
        tplot(
            "vec_xgse idl_thc_fgs_fac_xgse vecdiff",
            display=global_display,
            save_png=local_png,
        )

    def test_fac_xgse_thm(self):
        """Test of other_dim = xgse"""

        fac_matrix_make("idl_thc_fgs_gse_sm601", other_dim="xgse", newname="mat_xgse")
        tvector_rotate("mat_xgse", "idl_thc_fgs_gse", newname="vec_xgse")
        dat1 = get_data("mat_xgse")
        dat2 = get_data("vec_xgse")
        store_data("vecdiff", data={"x": dat2.times, "y": self.vec_xgse_thm.y - dat2.y})
        assert_allclose(dat1.y, self.mat_xgse_thm.y, atol=1.0e-06)
        assert_allclose(dat2.y, self.vec_xgse_thm.y, atol=1.0e-06)
        store_data("vecdiff", data={"x": dat2.times, "y": self.vec_xgse_thm.y - dat2.y})
        local_png = os.path.join(save_dir, "fac_xgse_thm.png")
        tplot(
            "vec_xgse idl_thc_fgs_fac_xgse vecdiff",
            display=global_display,
            save_png=local_png,
        )

    def test_fac_rgeo(self):
        """Test of other_dim = xgse"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="rgeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_rgeo",
        )
        tvector_rotate("mat_rgeo", "idl_thc_fgs_gse", newname="vec_rgeo")
        dat1 = get_data("mat_rgeo")
        dat2 = get_data("vec_rgeo")
        assert_allclose(dat1.y, self.mat_rgeo.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_rgeo.y, atol=0.002)

    def test_fac_rgeo_thm(self):
        """Test of other_dim = xgse"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="rgeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_rgeo",
        )
        tvector_rotate("mat_rgeo", "idl_thc_fgs_gse", newname="vec_rgeo")
        dat1 = get_data("mat_rgeo")
        dat2 = get_data("vec_rgeo")
        assert_allclose(dat1.y, self.mat_rgeo_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_rgeo_thm.y, atol=0.002)

    def test_fac_phigeo(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="phigeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_phigeo",
        )
        tvector_rotate("mat_phigeo", "idl_thc_fgs_gse", newname="vec_phigeo")
        dat1 = get_data("mat_phigeo")
        dat2 = get_data("vec_phigeo")
        assert_allclose(dat1.y, self.mat_phigeo.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_phigeo.y, atol=0.002)

    def test_fac_phigeo_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="phigeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_phigeo",
        )
        tvector_rotate("mat_phigeo", "idl_thc_fgs_gse", newname="vec_phigeo")
        dat1 = get_data("mat_phigeo")
        dat2 = get_data("vec_phigeo")
        assert_allclose(dat1.y, self.mat_phigeo_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_phigeo_thm.y, atol=0.002)

    def test_fac_mphigeo(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="mphigeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_mphigeo",
        )
        tvector_rotate("mat_mphigeo", "idl_thc_fgs_gse", newname="vec_mphigeo")
        dat1 = get_data("mat_mphigeo")
        dat2 = get_data("vec_mphigeo")
        assert_allclose(dat1.y, self.mat_mphigeo.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_mphigeo.y, atol=0.002)

    def test_fac_mphigeo_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="mphigeo",
            pos_var_name="idl_thc_state_pos",
            newname="mat_mphigeo",
        )
        tvector_rotate("mat_mphigeo", "idl_thc_fgs_gse", newname="vec_mphigeo")
        dat1 = get_data("mat_mphigeo")
        dat2 = get_data("vec_mphigeo")
        assert_allclose(dat1.y, self.mat_mphigeo_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_mphigeo_thm.y, atol=0.002)

    def test_fac_phism(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="phism",
            pos_var_name="idl_thc_state_pos",
            newname="mat_phism",
        )
        tvector_rotate("mat_phism", "idl_thc_fgs_gse", newname="vec_phism")
        dat1 = get_data("mat_phism")
        dat2 = get_data("vec_phism")
        assert_allclose(dat1.y, self.mat_phism.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_phism.y, atol=0.002)

    def test_fac_phism_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="phism",
            pos_var_name="idl_thc_state_pos",
            newname="mat_phism",
        )
        tvector_rotate("mat_phism", "idl_thc_fgs_gse", newname="vec_phism")
        dat1 = get_data("mat_phism")
        dat2 = get_data("vec_phism")
        assert_allclose(dat1.y, self.mat_phism_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_phism_thm.y, atol=0.002)

    def test_fac_mphism(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="mphism",
            pos_var_name="idl_thc_state_pos",
            newname="mat_mphism",
        )
        tvector_rotate("mat_mphism", "idl_thc_fgs_gse", newname="vec_mphism")
        dat1 = get_data("mat_mphism")
        dat2 = get_data("vec_mphism")
        assert_allclose(dat1.y, self.mat_mphism.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_mphism.y, atol=0.002)

    def test_fac_mphism_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="mphism",
            pos_var_name="idl_thc_state_pos",
            newname="mat_mphism",
        )
        tvector_rotate("mat_mphism", "idl_thc_fgs_gse", newname="vec_mphism")
        dat1 = get_data("mat_mphism")
        dat2 = get_data("vec_mphism")
        assert_allclose(dat1.y, self.mat_mphism_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_mphism_thm.y, atol=0.002)

    def test_fac_ygsm(self):
        """Test of other_dim = ygsm"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="ygsm",
            pos_var_name="idl_thc_state_pos",
            newname="mat_ygsm",
        )
        tvector_rotate("mat_ygsm", "idl_thc_fgs_gse", newname="vec_ygsm")
        dat1 = get_data("mat_ygsm")
        dat2 = get_data("vec_ygsm")
        assert_allclose(dat1.y, self.mat_ygsm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_ygsm.y, atol=0.002)

    def test_fac_ygsm_thm(self):
        """Test of other_dim = ygsm"""

        fac_matrix_make(
            "idl_thc_fgs_gse_sm601",
            other_dim="ygsm",
            pos_var_name="idl_thc_state_pos",
            newname="mat_ygsm",
        )
        tvector_rotate("mat_ygsm", "idl_thc_fgs_gse", newname="vec_ygsm")
        dat1 = get_data("mat_ygsm")
        dat2 = get_data("vec_ygsm")
        assert_allclose(dat1.y, self.mat_ygsm_thm.y, atol=0.002)
        assert_allclose(dat2.y, self.vec_ygsm_thm.y, atol=0.002)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot
import logging
import pyspedas
from pyspedas.tplot_tools import options

def ylim(name, min, max, logflag=None):
    """
    This function will set the y axis range displayed for a specific tplot variable.
    
    Parameters
    ----------
        name : str, int, list of str, list of int
            The names, indices, or wildcard patterns of the tplot variable that you wish to set y limits for.
        min : flt
            The start of the y axis.
        max : flt
            The end of the y axis.
        logflag: bool
            (Optional) If True, the y axis will be logarithmic scale, if False, linear scale, if None, no change
            
    Returns
    -------
        None
    
    Examples
    --------
        >>> # Change the y range of Variable1 
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> pyspedas.ylim('Variable1', 2, 4)

    """
    if logflag is None:
        options(name,'y_range',[min, max])
    else:
        options(name,'y_range',[min, max, logflag])
    return